#
#
#
#
#


import contextlib
import os
import platform
from pathlib import Path


def scan_dir(path, files_only=False, dirs_only=False, drop=[]):
    assert os.path.exists(path)
    checks = [
        lambda entry: not files_only or (files_only and entry.is_file()),
        lambda entry: not dirs_only or (dirs_only and entry.is_dir()),
        lambda entry: entry.name not in drop,
    ]
    return [
        entry.name
        for entry in os.scandir(path)
        if all([check(entry) for check in checks])
    ]


@contextlib.contextmanager
def pushd(new_cwd):
    if not os.path.exists(new_cwd):
        raise RuntimeError("phlop.os.pushd: new_cwd does not exist")

    cwd = os.getcwd()
    os.chdir(new_cwd)
    try:
        yield
    finally:
        os.chdir(cwd)


def write_to_file(file, contents, mode="w", skip_if_empty=True):
    if contents or not skip_if_empty:
        try:
            Path(file).parent.mkdir(parents=True, exist_ok=True)
            with open(file, mode) as f:
                f.write(contents)
        except IOError as e:
            raise RuntimeError(f"Failed to write to file {file}: {e}")


def env_sep():
    return ";" if any(platform.win32_ver()) else ":"



#
#
#
#
#

import subprocess

from phlop.os import pushd, write_to_file
from phlop.procs.runtimer import RunTimer
from phlop.string import decode_bytes


class ProcessNonZeroExitCode(RuntimeError):
    ...


def run(cmd, shell=True, capture_output=True, check=False, print_cmd=True, **kwargs):
    """https://docs.python.org/3/library/subprocess.html"""
    if print_cmd:
        print(f"running: {cmd}")

    return RunTimer(
        cmd, shell=shell, capture_output=capture_output, check=check, **kwargs
    )


def run_raw(args: list, shell=False, quiet=False):
    pipe = subprocess.DEVNULL if quiet else None
    return subprocess.Popen(
        args, shell=shell, stdin=pipe, stdout=pipe, stderr=pipe, close_fds=True
    )


def run_mp(cmds, N_CORES=None, **kwargs):
    """
    spawns N_CORES threads (default=len(cmds)) running commands and waiting for results
    https://docs.python.org/3/library/concurrent.futures.html
    """
    import concurrent.futures

    if N_CORES is None:
        N_CORES = len(cmds)

    with concurrent.futures.ThreadPoolExecutor(max_workers=N_CORES) as executor:
        jobs = [executor.submit(run, cmd, **kwargs) for cmd in cmds]
        results = []
        for future in concurrent.futures.as_completed(jobs):
            try:
                results += [future.result()]
                if future.exception() is not None:
                    raise future.exception()
            except Exception as exc:
                if kwargs.get("check", False):
                    executor.shutdown(wait=False, cancel_futures=True)
                    raise exc
                else:
                    print(f"run_mp generated an exception: {exc}")
        return results


def binary_exists_on_path(bin):
    """
    https://linux.die.net/man/1/which
    """
    raise ValueError("do better")
    return run(f"which {bin}").exitcode == 0



#
#
#
#
#

import importlib
import inspect
import logging
import os
from pathlib import Path

from phlop.sys import extend_sys_path

FORCE_RAISE_ON_IMPORT_ERROR = os.getenv(
    "PHLOP_FORCE_RAISE_ON_IMPORT_ERROR", "False"
).lower() in ("true", "1", "t")


def classes_in_file(file_path, subclasses_only=None, fail_on_import_error=False):
    file = Path(file_path)
    assert file.exists()
    module = str(file).replace(os.path.sep, ".")[:-3]
    assert module

    if subclasses_only is not None and not isinstance(subclasses_only, list):
        subclasses_only = [subclasses_only]

    classes = []

    with extend_sys_path([os.getcwd(), str(file.parent)]):
        try:
            for name, cls in inspect.getmembers(
                importlib.import_module(module), inspect.isclass
            ):
                should_add = subclasses_only == None or any(
                    [issubclass(cls, sub) for sub in subclasses_only]
                )
                if should_add:
                    classes += [cls]
        except (ValueError, ModuleNotFoundError) as e:
            if fail_on_import_error or FORCE_RAISE_ON_IMPORT_ERROR:
                raise e
            logging.error(f"Skipping on error: {e} in module {module}")

    return classes


def classes_in_directory(path, subclasses_only=None):
    classes = []
    for file in list(Path(path).glob("**/*.py")):
        classes += classes_in_file(file, subclasses_only)
    return classes



#
#
#
#
#


import logging
import os
import signal
import sys
import time
from dataclasses import dataclass, field
from datetime import datetime
from multiprocessing import Process, Queue

import numpy as np
import psutil
import yaml

from phlop.dict import ValDict
from phlop.proc import run_raw

logging.basicConfig(level=logging.DEBUG)
logger = logging.getLogger(__name__)

_default_interval = 2


@dataclass
class ProcessCaptureInfo:
    cpu_load: list = field(default_factory=lambda: [])
    fds: list = field(default_factory=lambda: [])
    mem_usage: list = field(default_factory=lambda: [])
    timestamps: list = field(default_factory=lambda: [])


def cli_args_parser():
    import argparse

    _help = ValDict(
        quiet="Redirect output to /dev/null",
        interval="Seconds between process stat capture",
    )
    parser = argparse.ArgumentParser()
    parser.add_argument("remaining", nargs=argparse.REMAINDER)
    parser.add_argument(
        "-q", "--quiet", action="store_true", default=False, help=_help.quiet
    )
    parser.add_argument(
        "-i", "--interval", default=_default_interval, help=_help.interval
    )
    return parser


def verify_cli_args(cli_args):
    try:
        cli_args.interval = int(cli_args.interval)
    except ValueError:
        raise ValueError("Interval must be an integer")
    sys.argv = [sys.argv[0]]  # drop everything!
    return cli_args


def check_pid(pid):
    """Check For the existence of a unix pid."""
    if not pid:
        return False
    try:
        os.kill(pid, 0)
    except OSError:
        return False
    return True


def signal_handler(sig, frame):
    sys.exit(0)


def bytes_as_mb(n_bytes):
    return int(n_bytes / 1024**2)


def capture_now(pid, data):
    now = datetime.utcnow().timestamp()
    proc = psutil.Process(pid=pid)
    data.cpu_load += [proc.cpu_percent(interval=0.1)]
    data.fds += [len(proc.open_files())]
    data.mem_usage += [bytes_as_mb(proc.memory_info().rss)]
    data.timestamps += [now]


class RuntimeStatsManager:
    def __init__(self, pid, interval=_default_interval):
        self.pid = pid
        self.interval = interval
        self.pqueue = Queue()
        self.data = {}
        self.p = Process(target=RuntimeStatsManager._run, args=(self,))
        self.p.start()
        self.data = self.pqueue.get()

    def __del__(self):
        if check_pid(self.p.pid):
            os.kill(self.p.pid, signal.SIGINT)
        self.join()

    def join(self):
        if not self.pid:
            return
        while self.p.exitcode is None and check_pid(self.p.pid):
            time.sleep(1)
        self.pid = 0
        return self

    @staticmethod
    def _run(this):
        data = ProcessCaptureInfo()
        signal.signal(signal.SIGINT, signal_handler)
        while check_pid(this.pid):
            try:
                capture_now(this.pid, data)
            except psutil.AccessDenied:
                break
            time.sleep(this.interval)
        this.pqueue.put(data)


def attach_to_this_process():
    return RuntimeStatsManager(os.getpid())


def print_summary(statsman):
    print("summary:")
    print("\tCPU avg:", np.average(statsman.data.cpu_load))
    print("\tCPU max:", np.max(statsman.data.cpu_load))

    print("\tFDS avg:", np.average(statsman.data.fds))
    print("\tFDS max:", np.max(statsman.data.fds))

    print("\tMEM avg:", np.average(statsman.data.mem_usage))
    print("\tMEM max:", np.max(statsman.data.mem_usage))


def main():
    parser = cli_args_parser()
    cli_args = verify_cli_args(parser.parse_args())
    try:
        proc = run_raw(cli_args.remaining, quiet=cli_args.quiet)
        statsman = RuntimeStatsManager(proc.pid, cli_args.interval).join()
        print_summary(statsman)
    except (Exception, SystemExit) as e:
        logger.exception(e)
        parser.print_help()
    except:
        e = sys.exc_info()[0]
        print(f"Error: Unknown Error {e}")


if __name__ == "__main__":
    main()



#
#
#
#
#


import os
import subprocess
import time
from copy import deepcopy

from phlop.os import pushd, write_to_file
from phlop.string import decode_bytes


class RunTimer:
    def __init__(
        self,
        cmd,
        shell=False,
        capture_output=True,
        check=False,
        print_cmd=True,
        env: dict = {},  # dict[str, str] # eventually
        working_dir=None,
        log_file_path=None,
        logging=2,
        **kwargs,
    ):
        self.cmd = cmd
        start = time.time()

        self.run_time = None
        self.stdout = ""
        self.stderr = ""
        benv = os.environ.copy()
        benv.update(env)
        self.logging = logging

        ekwargs = {}
        if not capture_output and log_file_path:
            ekwargs.update(
                dict(
                    stdout=open(f"{log_file_path}.stdout", "w"),
                    stderr=open(f"{log_file_path}.stderr", "w"),
                ),
            )

        def run():
            logging = deepcopy(self.logging)
            try:
                self.run = subprocess.run(
                    self.cmd,
                    shell=shell,
                    check=check,
                    env=benv,
                    capture_output=capture_output,
                    close_fds=True,
                    **kwargs,
                    **ekwargs,
                )
                self.run_time = time.time() - start
                self.exitcode = self.run.returncode
                if logging == 2 and capture_output:
                    self.stdout = decode_bytes(self.run.stdout)
                    self.stderr = decode_bytes(self.run.stderr)
            except (
                subprocess.CalledProcessError
            ) as e:  # only triggers on failure if check=True
                self.exitcode = e.returncode
                self.run_time = time.time() - start
                if logging >= 1 and capture_output:
                    self.stdout = decode_bytes(e.stdout)
                    self.stderr = decode_bytes(e.stderr)
                    logging = 2  # force logging as exception occurred
            if logging == 2 and capture_output and log_file_path:
                write_to_file(f"{log_file_path}.stdout", self.stdout)
                write_to_file(f"{log_file_path}.stderr", self.stderr)

        if working_dir:
            with pushd(working_dir):
                run()
        else:
            run()

    def out(self, ignore_exit_code=False):
        if not ignore_exit_code and self.exitcode > 0:
            raise RuntimeError(f"phlop.RunTimer error: {self.stderr}")
        return self.stdout
